==> CRTL-0.8.3-py3-none-any.whl: crtl/components/crtl_el.py <==
import itertools
import pathlib
from typing import Callable, Iterable, List, Optional, Union
from spacy.tokens import Doc, Span,SpanGroup
from spacy.util import ensure_path

from crtl.components.ann_vector_kb import ANNCandidateGenerator, MentionCandidate
from crtl.lkb.lkb import LexicalKnowledgeBase
from spacy.language import Language
from spacy.pipeline import TrainablePipe

class CRTLTextLinker:
    NIL = "NIL"

    def __init__(self, candidate_generator, threshold, ef_search):
        self._threshold = threshold
        self._ef_search = ef_search
        self._candidate_generator = candidate_generator

    def predict(self, mention_strings: Iterable[str]) -> Union[List[str], List[int]]:
        batch_candidates = self._candidate_generator(mention_strings,k=1)
        return self._get_winning_links(batch_candidates)

    def _get_winning_links(
        self, batch_candidates: Iterable[Iterable[MentionCandidate]]
    ) -> Union[List[str], List[int]]:  # can pass this function in to allow more customizabiliity?
        concept_ids = []
        for candidates in batch_candidates:
            try:
                candidate = next(iter(candidates))
                if candidate.similarity > self._threshold:
                    concept_ids.append(candidate.concept_id)
                else:
                    concept_ids.append(self.NIL)
            except:
                concept_ids.append(self.NIL)
        return concept_ids

    def to_disk(self, path: pathlib.Path):
        # This will receive the directory path + /my_component
        path = ensure_path(path)
        if not path.exists():
            path.mkdir()
        ann_candidate_generator_path = path / "candidate_generator"
        self._candidate_generator.to_disk(ann_candidate_generator_path)

    @classmethod
    def from_disk(cls, path: pathlib.Path, threshold, ef_search):
        path = ensure_path(path)
        if not path.exists():
            path.mkdir()
        # This will receive the directory path + /my_component
        ann_candidate_generator_path = path / "candidate_generator"
        candidate_generator = ANNCandidateGenerator.from_disk(
            ann_candidate_generator_path
        )
        # how do we make this use an abstract candidate generator
        return cls(candidate_generator, threshold, ef_search)


class CRTLLinker(TrainablePipe):
    def __init__(self, threshold: int, ef_search: int,scorer:Optional[Callable]=None):
        self.scorer =scorer
        self._threshold = threshold
        self._ef_search = ef_search

    def initialize(self, get_examples=None, nlp=None, candidate_generator_getter: Optional[Callable[[],ANNCandidateGenerator]]=None):

        cg = candidate_generator_getter()
        for concept_id in cg.lkb.cui_to_concept.keys():
            nlp.vocab.strings.add(concept_id)
        self._text_linker = CRTLTextLinker(
            cg, self._threshold, self._ef_search
        )

    def predict(self, docs: Iterable[Doc]) -> Union[List[str], List[int]]:
        raise NotImplementedError

    def set_annotations(
        self, docs: Iterable[Doc], concept_id_predictions: List[str]
    ) -> None:
        raise NotImplementedError

    def to_disk(self, path: pathlib.Path, exclude=tuple()):
        self._text_linker.to_disk(path)

    def from_disk(self, path: pathlib.Path, exclude=tuple()):
        self._text_linker = CRTLTextLinker.from_disk(
            path, self._threshold, self._ef_search
        )
        return self
    @property
    def is_trainable(self) -> bool:
        return False


class CRTLSpanLinker(CRTLLinker):
    def __init__(self, span_key,*args,**kwargs):
        super().__init__(*args)
        self._span_key = span_key

    def predict(self, docs: Iterable[Doc]) -> Union[List[str], List[int]]:
        mention_strings = itertools.chain.from_iterable(
            ((span.text for span in doc.spans[self._span_key]) for doc in docs)
        )
        return self._text_linker.predict(mention_strings)

    def set_annotations(
        self, docs: Iterable[Doc], concept_id_predictions: Union[List[str], List[int]]
    ) -> None:
        offset = 0 
        for doc in docs:
            spans = doc.spans[self._span_key]
            doc_concept_id_predictions = concept_id_predictions[offset:offset+len(spans)]
            offset += len(spans)
            new_spans = []
            for span, concept_id_prediction in zip(spans, doc_concept_id_predictions):
                new_spans.append( Span(doc, span.start, span.end,label=span.label_,kb_id=str(concept_id_prediction)) )
            doc.spans[self._span_key] = SpanGroup(doc,name=self._span_key,spans=new_spans)
            print([span.kb_id_ for span in doc.spans[self._span_key]])

class CRTLEntityLinker(CRTLLinker):
    def predict(self, docs: Iterable[Doc]) -> Union[List[str], List[int]]:
        mention_strings = itertools.chain.from_iterable(
            ((ent.text for ent in doc.ents) for doc in docs)
        )
        return self._text_linker.predict(mention_strings)

    def set_annotations(
        self, docs: Iterable[Doc], concept_id_predictions: Union[List[str], List[int]]
    ) -> None:
        ents = itertools.chain.from_iterable(doc.ents for doc in docs)
        for ent, concept_id_prediction in zip(ents, concept_id_predictions):
            for token in ent:
                token.ent_kb_id_ = str(concept_id_prediction)


@Language.factory(
    name="trend_ngram_tfidf_entity_linker",
    default_config={"threshold": 0.7, 
                    "ef_search": 200, 
                    "scorer": {"@scorers": "spacy.entity_linker_scorer.v1"}},
    default_score_weights={
        "nel_micro_f": 1.0,
        "nel_micro_r": None,
        "nel_micro_p": None,
    },
)
def make_trend_ngram_tfidf_entity_linker(
    nlp: Language, name: str, threshold: int, ef_search: int,scorer:Optional[Callable]
):
    return CRTLEntityLinker(threshold, ef_search,scorer)


@Language.factory(
    name="trend_ngram_tfidf_span_linker",
    default_config={"threshold": 0.7, 
                    "ef_search": 200, 
                    "span_key": "sc",
                    "scorer": {"@scorers": "crtl.score_span_links","span_key":"sc"}},
    default_score_weights={
        "nel_micro_f": 1.0,
        "nel_micro_r": None,
        "nel_micro_p": None,
    },
)
def make_trend_ngram_tfidf_span_linker(
    nlp: Language, name: str, threshold: int, ef_search: int, span_key: str,scorer:Optional[Callable]
):
    return CRTLSpanLinker(span_key, threshold, ef_search,scorer)

==> CRTL-0.11.0-py3-none-any.whl: crtl/components/crtl_el.py <==
import itertools
import pathlib
from typing import Callable, Iterable, List, Optional, Union
from spacy.tokens import Doc, Span,SpanGroup
from spacy.util import ensure_path

from crtl.components.ann_vector_kb import ANNCandidateGenerator, MentionCandidate
from crtl.lkb.lkb import LexicalKnowledgeBase
from spacy.language import Language
from spacy.pipeline import TrainablePipe

class CRTLTextLinker:
    NIL = "NIL"

    def __init__(self, candidate_generator, threshold, ef_search):
        self._threshold = threshold
        self._ef_search = ef_search
        self._candidate_generator = candidate_generator

    def predict(self, mention_strings: Iterable[str]) -> Union[List[str], List[int]]:
        batch_candidates = self._candidate_generator(mention_strings,k=1)
        return self._get_winning_links(batch_candidates)

    def _get_winning_links(
        self, batch_candidates: Iterable[Iterable[MentionCandidate]]
    ) -> Union[List[str], List[int]]:  # can pass this function in to allow more customizabiliity?
        concept_ids = []
        for candidates in batch_candidates:
            try:
                candidate = next(iter(candidates))
                if candidate.similarity > self._threshold:
                    concept_ids.append(candidate.concept_id)
                else:
                    concept_ids.append(self.NIL)
            except:
                concept_ids.append(self.NIL)
        return concept_ids

    def to_disk(self, path: pathlib.Path):
        # This will receive the directory path + /my_component
        path = ensure_path(path)
        if not path.exists():
            path.mkdir()
        ann_candidate_generator_path = path / "candidate_generator"
        self._candidate_generator.to_disk(ann_candidate_generator_path)

    @classmethod
    def from_disk(cls, path: pathlib.Path, threshold, ef_search):
        path = ensure_path(path)
        if not path.exists():
            path.mkdir()
        # This will receive the directory path + /my_component
        ann_candidate_generator_path = path / "candidate_generator"
        candidate_generator = ANNCandidateGenerator.from_disk(
            ann_candidate_generator_path
        )
        # how do we make this use an abstract candidate generator
        return cls(candidate_generator, threshold, ef_search)


class CRTLLinker(TrainablePipe):
    def __init__(self, threshold: int, ef_search: int,scorer:Optional[Callable]=None):
        self.scorer =scorer
        self._threshold = threshold
        self._ef_search = ef_search

    def initialize(self, get_examples=None, nlp=None, candidate_generator_getter: Optional[Callable[[],ANNCandidateGenerator]]=None):

        cg = candidate_generator_getter()
        for concept_id in cg.lkb.cui_to_concept.keys():
            nlp.vocab.strings.add(concept_id)
        self._text_linker = CRTLTextLinker(
            cg, self._threshold, self._ef_search
        )

    def predict(self, docs: Iterable[Doc]) -> Union[List[str], List[int]]:
        raise NotImplementedError

    def set_annotations(
        self, docs: Iterable[Doc], concept_id_predictions: List[str]
    ) -> None:
        raise NotImplementedError

    def to_disk(self, path: pathlib.Path, exclude=tuple()):
        self._text_linker.to_disk(path)

    def from_disk(self, path: pathlib.Path, exclude=tuple()):
        self._text_linker = CRTLTextLinker.from_disk(
            path, self._threshold, self._ef_search
        )
        return self
    @property
    def is_trainable(self) -> bool:
        return False


class CRTLSpanLinker(CRTLLinker):
    def __init__(self, span_key,*args,**kwargs):
        super().__init__(*args)
        self._span_key = span_key

    def predict(self, docs: Iterable[Doc]) -> Union[List[str], List[int]]:
        mention_strings = itertools.chain.from_iterable(
            ((span.text for span in doc.spans[self._span_key]) for doc in docs)
        )
        return self._text_linker.predict(mention_strings)

    def set_annotations(
        self, docs: Iterable[Doc], concept_id_predictions: Union[List[str], List[int]]
    ) -> None:
        offset = 0 
        for doc in docs:
            spans = doc.spans[self._span_key]
            doc_concept_id_predictions = concept_id_predictions[offset:offset+len(spans)]
            offset += len(spans)
            new_spans = []
            for span, concept_id_prediction in zip(spans, doc_concept_id_predictions):
                new_spans.append( Span(doc, span.start, span.end,label=span.label_,kb_id=str(concept_id_prediction)) )
            doc.spans[self._span_key] = SpanGroup(doc,name=self._span_key,spans=new_spans)

class CRTLEntityLinker(CRTLLinker):
    def predict(self, docs: Iterable[Doc]) -> Union[List[str], List[int]]:
        mention_strings = itertools.chain.from_iterable(
            ((ent.text for ent in doc.ents) for doc in docs)
        )
        return self._text_linker.predict(mention_strings)

    def set_annotations(
        self, docs: Iterable[Doc], concept_id_predictions: Union[List[str], List[int]]
    ) -> None:
        ents = itertools.chain.from_iterable(doc.ents for doc in docs)
        for ent, concept_id_prediction in zip(ents, concept_id_predictions):
            for token in ent:
                token.ent_kb_id_ = str(concept_id_prediction)


@Language.factory(
    name="trend_ngram_tfidf_entity_linker",
    default_config={"threshold": 0.7, 
                    "ef_search": 200, 
                    "scorer": {"@scorers": "crtl.score_ent_links"}},
    default_score_weights={
        "nel_micro_f": 1.0,
        "nel_micro_r": None,
        "nel_micro_p": None,
    },
)
def make_trend_ngram_tfidf_entity_linker(
    nlp: Language, name: str, threshold: int, ef_search: int,scorer:Optional[Callable]
):
    return CRTLEntityLinker(threshold, ef_search,scorer)


@Language.factory(
    name="trend_ngram_tfidf_span_linker",
    default_config={"threshold": 0.7, 
                    "ef_search": 200, 
                    "span_key": "sc",
                    "scorer": {"@scorers": "crtl.score_span_links","span_key":"sc"}},
    default_score_weights={
        "strong_nel_ete_f": 1.0,
    },
)
def make_trend_ngram_tfidf_span_linker(
    nlp: Language, name: str, threshold: int, ef_search: int, span_key: str,scorer:Optional[Callable]
):
    return CRTLSpanLinker(span_key, threshold, ef_search,scorer)
